# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import units
import sushy

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import inspect
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

INFO_DICT = db_utils.get_test_redfish_info()


class MockedSushyError(Exception):
    pass


class RedfishInspectTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishInspectTestCase, self).setUp()
        self.config(enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-virtual-media'],
                    enabled_management_interfaces=['redfish'],
                    enabled_inspect_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT,
            provision_state=states.MANAGEABLE)

    def init_system_mock(self, system_mock, **properties):

        system_mock.reset()

        system_mock.boot.mode = sushy.BOOT_SOURCE_MODE_UEFI

        system_mock.memory_summary.size_gib = 2

        mock_processor = mock.Mock(
            spec=sushy.resources.system.processor.Processor)
        mock_processor.model = 'test'
        mock_processor.processor_architecture = sushy.PROCESSOR_ARCH_x86
        mock_processor.instruction_set = sushy.InstructionSet.X86_64
        mock_processor.max_speed_mhz = 1234
        mock_processor.total_threads = 8
        system_mock.processors.get_members.return_value = [mock_processor]

        # make the summary follow the data above by making it
        # a property like sushy and returning the same data
        type(system_mock.processors).summary = mock.PropertyMock(
            side_effect=lambda:
                sushy.resources.system.processor.ProcessorSummary(
                    count=mock_processor.total_threads,
                    architecture=mock_processor.processor_architecture
                )
        )

        mock_storage_drive = mock.Mock(
            spec=sushy.resources.system.storage.drive.Drive)
        mock_storage_drive.name = 'storage-drive'
        mock_storage_drive.capacity_bytes = '128'

        mock_storage = mock.Mock(
            spec=sushy.resources.system.storage.storage.Storage)
        mock_storage.drives = [mock_storage_drive]
        system_mock.storage.get_members.return_value = [
            mock_storage]

        mock_simple_storage_device = mock.Mock(
            spec=sushy.resources.system.simple_storage.DeviceListField)
        mock_simple_storage_device.name = 'test-name'
        mock_simple_storage_device.capacity_bytes = '123'

        mock_simple_storage = mock.Mock(
            spec=sushy.resources.system.simple_storage.SimpleStorage)
        mock_simple_storage.devices = [mock_simple_storage_device]
        system_mock.simple_storage.get_members.return_value = [
            mock_simple_storage]

        system_mock.simple_storage.disks_sizes_bytes = (
            1 * units.Gi, units.Gi * 3, units.Gi * 5)
        system_mock.storage.volumes_sizes_bytes = (
            2 * units.Gi, units.Gi * 4, units.Gi * 6)

        eth_interface_mock1 = mock.Mock(
            spec=sushy.resources.system.ethernet_interface.EthernetInterface)
        eth_interface_mock1.identity = 'NIC.Integrated.1-1'
        eth_interface_mock1.mac_address = '00:11:22:33:44:55'
        eth_interface_mock1.status.state = sushy.STATE_ENABLED
        eth_interface_mock1.status.health = sushy.HEALTH_OK

        eth_interface_mock2 = mock.Mock(
            spec=sushy.resources.system.ethernet_interface.EthernetInterface)
        eth_interface_mock2.identity = 'NIC.Integrated.2-1'
        eth_interface_mock2.mac_address = '66:77:88:99:AA:BB'
        eth_interface_mock2.status.state = sushy.STATE_DISABLED
        eth_interface_mock2.status.health = sushy.HEALTH_OK

        system_mock.ethernet_interfaces.get_members.return_value = [
            eth_interface_mock1,
            eth_interface_mock2
        ]

        # make the summary follow the data above by making it
        # a property like sushy and returning the same data
        type(system_mock.ethernet_interfaces).summary = mock.PropertyMock(
            side_effect=lambda: {
                obj.mac_address: obj.status.state
                for obj in
                system_mock.ethernet_interfaces.get_members.return_value
            }
        )

        system_mock.name = 'System1'

        system_mock.serial_number = '123456'

        system_mock.manufacturer = 'Sushy Emulator'

        return system_mock

    def test_get_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            properties = task.driver.get_properties()
            for prop in redfish_utils.COMMON_PROPERTIES:
                self.assertIn(prop, properties)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.management.validate(task)
            mock_parse_driver_info.assert_called_once_with(task.node)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    def test_inspect_hardware_ok(self, mock_create_ports_if_not_exist,
                                 mock_get_system,
                                 mock_get_enabled_macs):
        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': 2048,
        }
        self.init_system_mock(mock_get_system.return_value)

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)
            self.assertEqual(1, mock_create_ports_if_not_exist.call_count)
            mock_get_system.assert_called_once_with(task.node)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)

        system_vendor = inventory['inventory']['system_vendor']
        expected_product_name = 'System1'
        expected_serial_number = '123456'
        expected_manufacturer = 'Sushy Emulator'
        self.assertEqual(expected_product_name,
                         system_vendor['product_name'])
        self.assertEqual(expected_serial_number,
                         system_vendor['serial_number'])
        self.assertEqual(expected_manufacturer,
                         system_vendor['manufacturer'])

        expected_interfaces = [{'mac_address': '00:11:22:33:44:55',
                                'name': 'NIC.Integrated.1-1'},
                               {'mac_address': '66:77:88:99:AA:BB',
                                'name': 'NIC.Integrated.2-1'}]
        self.assertEqual(expected_interfaces,
                         inventory['inventory']['interfaces'])

        expected_cpu = {'count': 8, 'model_name': 'test',
                        'frequency': 1234, 'architecture': 'x86_64'}
        self.assertEqual(expected_cpu,
                         inventory['inventory']['cpu'])

        expected_disks = [{'name': 'storage-drive', 'size': '128'}]
        self.assertEqual(expected_disks,
                         inventory["inventory"]['disks'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    def test_inspect_port_creation(self, mock_create_ports_if_not_exist,
                                   mock_get_system,
                                   mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)
            result = task.driver.management.get_mac_addresses(task)
            inspect_utils.create_ports_if_not_exist.assert_called_once_with(
                task, result)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_fail_missing_cpu_arch(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_processor = system_mock.processors.get_members.return_value[0]
        mock_processor.processor_architecture = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.properties.pop('cpu_arch')
            # self.assertRaises(exception.HardwareInspectionFailure,
            #                  task.driver.inspect.inspect_hardware, task)
            # TODO(cardoe):
            # not a valid test currently because the normal inspection
            # path requires that architecture is filled out and populates
            # any value so this passes so need to reconcile that behavior
            # difference and return here.
            task.driver.inspect.inspect_hardware(task)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_cpu_count(self, mock_get_system,
                                                       mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_processor = system_mock.processors.get_members.return_value[0]
        mock_processor.total_threads = 0

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64', 'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertIn('count', inventory['inventory']['cpu'])
        self.assertEqual(0, inventory['inventory']['cpu']['count'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_cpu_model(self, mock_get_system,
                                                       mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_processor = system_mock.processors.get_members.return_value[0]
        mock_processor.model = None

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('model', inventory['inventory']['cpu'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_cpu_frequency(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_processor = system_mock.processors.get_members.return_value[0]
        mock_processor.max_speed_mhz = None

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('frequency', inventory['inventory']['cpu'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_cpu_instruction_set(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_processor = system_mock.processors.get_members.return_value[0]
        mock_processor.instruction_set = None

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': '',
            'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertIn('architecture', inventory['inventory']['cpu'])
        self.assertEqual('', inventory['inventory']['cpu']['architecture'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_local_gb(self, mock_get_system,
                                                      mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.simple_storage.disks_sizes_bytes = None
        system_mock.storage.volumes_sizes_bytes = None

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '0', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_simple_storage_and_storage(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.simple_storage = {}
        system_mock.storage = {}

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '0', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('disks', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_simple_storage(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.simple_storage = {}

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertIn('disks', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_storage(self, mock_get_system,
                                                     mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.storage = {}

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '4', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertIn('disks', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_fail_missing_memory_mb(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.memory_summary.size_gib = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.properties.pop('memory_mb')
            self.assertRaises(exception.HardwareInspectionFailure,
                              task.driver.inspect.inspect_hardware, task)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_memory_mb(self, mock_get_system,
                                                       mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.memory_summary.size_gib = None

        expected_properties = {
            'capabilities': 'boot_mode:uefi',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': '4096'
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('memory', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    def test_inspect_hardware_ignore_missing_nics(
            self, mock_create_ports_if_not_exist, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.ethernet_interfaces.get_members.return_value = []

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.HardwareInspectionFailure,
                              task.driver.inspect.inspect_hardware, task)
            self.assertFalse(mock_create_ports_if_not_exist.called)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_cpus(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.processors = []

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertIn('cpu', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_preserve_boot_mode(self, mock_get_system,
                                                 mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        expected_properties = {
            'capabilities': 'boot_mode:bios',
            'cpu_arch': 'x86_64',
            'local_gb': '3', 'memory_mb': 2048
        }

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.properties = {
                'capabilities': 'boot_mode:bios'
            }

            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        expected_boot_mode = {'current_boot_mode': 'uefi'}
        self.assertEqual(expected_boot_mode,
                         inventory['inventory']['boot'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_boot_mode(self, mock_get_system,
                                                       mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            expected_properties = {
                'cpu_arch': 'x86_64',
                'local_gb': '3', 'memory_mb': 2048
            }
            task.driver.inspect.inspect_hardware(task)

        self.node.refresh()
        self.assertEqual(expected_properties, self.node.properties)
        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('boot', inventory['inventory'])

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(objects.Port, 'list_by_node_id') # noqa
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_set_port_pxe_enabled(
            self, mock_get_system,
            mock_list_by_node_id,
            mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        pxe_disabled_port = obj_utils.create_test_port(
            self.context, uuid=self.node.uuid, node_id=self.node.id,
            address='00:11:22:33:44:55', pxe_enabled=False)
        mock_list_by_node_id.return_value = [pxe_disabled_port]
        port = mock_list_by_node_id.return_value
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs = mock.Mock()
            task.driver.inspect._get_pxe_port_macs.return_value = \
                ['00:11:22:33:44:55']
            task.driver.inspect.inspect_hardware(task)
            self.assertTrue(port[0].pxe_enabled)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(objects.Port, 'list_by_node_id') # noqa
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_set_port_pxe_disabled(
            self, mock_get_system,
            mock_list_by_node_id, mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        pxe_enabled_port = obj_utils.create_test_port(
            self.context, uuid=self.node.uuid,
            node_id=self.node.id, address='00:11:22:33:44:55',
            pxe_enabled=True)
        mock_list_by_node_id.return_value = [pxe_enabled_port]

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs = mock.Mock()
            task.driver.inspect._get_pxe_port_macs.return_value = \
                []
            task.driver.inspect.inspect_hardware(task)
            port = mock_list_by_node_id.return_value
            self.assertFalse(port[0].pxe_enabled)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(objects.Port, 'list_by_node_id') # noqa
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_conf_update_pxe_disabled_false(
            self, mock_get_system,
            mock_list_by_node_id, mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)
        pxe_enabled_port = obj_utils.create_test_port(
            self.context, uuid=self.node.uuid,
            node_id=self.node.id, address='00:11:22:33:44:55',
            pxe_enabled=False)
        mock_list_by_node_id.return_value = [pxe_enabled_port]

        self.config(update_pxe_enabled=False, group='inspector')

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs = mock.Mock()
            task.driver.inspect._get_pxe_port_macs.return_value = \
                ['00:11:22:33:44:55']
            task.driver.inspect.inspect_hardware(task)
            port = mock_list_by_node_id.return_value
            self.assertFalse(port[0].pxe_enabled)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_no_mac(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        mock_eths = system_mock.ethernet_interfaces.get_members.return_value
        mock_eths[1].mac_address = ''
        mock_eths[1].status.state = sushy.STATE_ENABLED

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)
            ports = objects.Port.list_by_node_id(task.context, self.node.id)
            self.assertEqual(1, len(ports))

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(objects.Port, 'list_by_node_id') # noqa
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_empty_pxe_port_macs(
            self, mock_get_system,
            mock_list_by_node_id, mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        pxe_enabled_port = obj_utils.create_test_port(
            self.context, uuid=self.node.uuid,
            node_id=self.node.id, address='24:6E:96:70:49:01',
            pxe_enabled=True)
        mock_list_by_node_id.return_value = [pxe_enabled_port]

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs = mock.Mock()
            task.driver.inspect._get_pxe_port_macs.return_value = []
            return_value = task.driver.inspect.inspect_hardware(task)
            port = mock_list_by_node_id.return_value
            self.assertFalse(port[0].pxe_enabled)
            self.assertEqual(states.MANAGEABLE, return_value)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(objects.Port, 'list_by_node_id') # noqa
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_with_none_pxe_port_macs(
            self, mock_get_system,
            mock_list_by_node_id, mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)

        pxe_enabled_port = obj_utils.create_test_port(
            self.context, uuid=self.node.uuid,
            node_id=self.node.id, address='00:11:22:33:44:55',
            pxe_enabled=True)
        mock_list_by_node_id.return_value = [pxe_enabled_port]

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs = mock.Mock()
            task.driver.inspect._get_pxe_port_macs.return_value = None
            task.driver.inspect.inspect_hardware(task)
            port = mock_list_by_node_id.return_value
            self.assertFalse(port[0].pxe_enabled)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_create_port_when_its_state_is_none(self, mock_get_system,
                                                mock_get_enabled_macs):
        self.init_system_mock(mock_get_system.return_value)
        expected_port_mac_list = ["00:11:22:33:44:55", "66:77:88:99:aa:bb"]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)
            ports = objects.Port.list_by_node_id(task.context, self.node.id)
            for port in ports:
                self.assertIn(port.address, expected_port_mac_list)

    @mock.patch.object(redfish_utils, 'get_enabled_macs', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_inspect_hardware_ignore_missing_system_vendor(
            self, mock_get_system, mock_get_enabled_macs):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.name = None
        system_mock.serial_number = None
        system_mock.manufacturer = None

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect.inspect_hardware(task)

        inventory = inspect_utils.get_inspection_data(self.node,
                                                      self.context)
        self.assertNotIn('system_vendor', inventory['inventory'])

    def test_get_pxe_port_macs(self):
        expected_properties = None
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_properties,
                             task.driver.inspect._get_pxe_port_macs(task))


class ContinueInspectionTestCase(db_base.DbTestCase):
    def setUp(self):
        super(ContinueInspectionTestCase, self).setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['redfish', 'no-inspect'])
        self.config(inspection_hooks='validate-interfaces,'
                                     'ports,architecture',
                    group='redfish')
        self.config(enabled_hardware_types=['redfish'])
        self.config(enabled_power_interfaces=['redfish'])
        self.config(enabled_management_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context,
            driver='redfish',
            inspect_interface='redfish',
            driver_info=INFO_DICT,
            provision_state=states.INSPECTING)
        self.iface = inspect.RedfishInspect()

    def test(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(
                exception.UnsupportedDriverExtension,
                self.iface.continue_inspection,
                task,
                mock.sentinel.inventory,
                mock.sentinel.plugin_data
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import gzip
import json
import os
import tempfile

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
import pycdlib

from ironic.common import exception
from ironic.common import neutron


LOG = log.getLogger(__name__)
CONF = cfg.CONF


def is_invalid_network_metadata(network_data):
    """Return True if network metadata is invalid.

    :param network_data: A dictionary object containing network_metadata,
                         with three sub-fields, 'links', 'networks',
                         and 'services'. Returns True if the document is
                         invalid and missing network metadata, or when
                         specific other issues such as null MTU values
                         are detected.
    :returns: True when the data supplied is invalid.
    """
    try:
        # FIXME(TheJulia): Another possible issue is Nova can send
        # the MTU on links as "null" values. That seems super problematic
        # for jumbo frame users or constrained environments. We should
        # likely just declare missing MTU as a qualifier to rebuild, but
        # that is likely to then trigger far more often. Maybe that means
        # it is an even better idea...
        if (not network_data.get('links', [])
            and not network_data.get('networks', [])
            and not network_data.get('services', [])):
            return True
        if not CONF.conductor.disable_metadata_mtu_check:
            for link in network_data.get('links', []):
                if link.get('mtu') is None:
                    return True
    except AttributeError:
        # Got called with None or something else lacking the attribute.
        # NOTE(TheJulia): If we ever want to inject metadata if missing
        # here is where we would add it.
        return True
    return False


def check_and_patch_configdrive(task, configdrive):
    """Evaluates and patches an provided ISO9660 configuration drive file.

    This method is responsible for taking the user requested configuration
    drive, performing basic evaluation, attempting to patch/fix the
    configuration drive to enable the latest metadata written to be valid and
    correct, which is intended to allow the overall deployment process to
    proceed with the correct data.

    :param task: A TaskManager object.
    :param configdrive: The in-memory configuration drive file which was
                        submitted by the requester.
    :returns: The submitted configuration drive or a regenerated and patched
              configuration drive OR the original requested configuration
              drive *if* a failure is encountered in the overall process.
    """
    try:
        with tempfile.NamedTemporaryFile(dir=CONF.tempdir,
                                         mode="wb+") as fileobj:
            # We have a file, we need to extract it and get it to an temp file
            fileobj.write(gzip.decompress(base64.b64decode(configdrive)))
            fileobj.flush()
            # Reset the position back to the start.
            fileobj.seek(0)
            # We then needs to peak inside, and likely then re-master it.
            provided_iso = pycdlib.PyCdlib()
            provided_iso.open_fp(fileobj)
            # NOTE(TheJulia): on the filename is an ISO9660 detail which
            # delineates the filename as the end of filename and signals the
            # the file identifier number.
            # https://wiki.osdev.org/ISO_9660
            with provided_iso.open_file_from_iso(
                    iso_path='/openstack/latest/network_data.json;1') as ndfp:
                try:
                    nd_dict = json.loads(ndfp.read())
                    # NOTE(TheJulia): This is the area of the code we would
                    # want to inject any additional checks in for patching in.
                    if not is_invalid_network_metadata(nd_dict):
                        # We know the config drive is okay, we can return what
                        # was submitted.
                        return configdrive
                except (TypeError, json.decoder.JSONDecodeError) as e:
                    # NOTE(TheJulia): We've gotten some sort of invalid data
                    # which is a possible case. So... just live with it and
                    # move on.
                    LOG.debug('Encountered error while trying to parse '
                              'configuration drive network_data.json file '
                              'for node UUID %(node)s, instance %(inst)s.'
                              'content which was submitted to Ironic. '
                              'Error: %(error)s',
                              {'error': e,
                               'node': task.node.uuid,
                               'inst': task.node.instance_uuid})
            LOG.warning('The provided metadata for the deployment of '
                        'node %s appears invalid, and will be '
                        'regenerated based upon the available '
                        'metadata.', task.node.uuid)
            network_data = json.dumps(generate_config_metadata(task))
            # Grab the filename to use. This is not with a context manager
            # because we are going to do the write-out of the new file
            # using mkisofs.
            new_iso_file = tempfile.mkstemp(dir=CONF.tempdir)[1]
            # NOTE(TheJulia): This is where we would make a decision to
            # inject any additional override files to the image as the
            # regenerate_iso call takes overrides.
            regenerate_iso(
                provided_iso, new_iso_file,
                {'/openstack/latest/network_data.json': network_data},
                node_uuid=task.node.uuid)
            new_drive = _read_config_drive(new_iso_file)
            # Remove the file.
            os.remove(new_iso_file)
    except (OSError, processutils.ProcessExecutionError) as e:
        # Catch any general OSError (permission issue) or
        # ProcessExecutionError from regenerate_iso.
        LOG.error('Failed to regenerate the configuration drive ISO '
                  'for node %(node)s, '
                  'roceeding with submitted metadata: %s',
                  {'error': e,
                   'node': task.node.uuid})
        return configdrive
    except exception.ConfigDriveRegenerationFailure:
        # If this is being raised, the generation failed. That should have
        # been logged where it was raised.
        return configdrive
    except pycdlib.pycdlibexception.PyCdlibInvalidInput:
        # File is missing in ISO or could not be found.
        LOG.error('Failed to process metadata, as the supplied configuration '
                  'drive ISO for the deployment of node %s lacked the '
                  'openstack/latest/network_data.json file.', task.node.uuid)
        return configdrive
    return new_drive


def _read_config_drive(file):
    # NOTE(TheJulia): This is minimal to keep the file open interaction out
    # of the unit test code path, with the goal of just keeping things a bit
    # more simple as well.
    with open(file, "rb") as new_iso:
        return base64.b64encode(
            gzip.compress(new_iso.read())).decode()


def regenerate_iso(source, dest, override_files, node_uuid=None):
    """Utility method to regenerate a config drive ISO image.

    This method takes the source, extracts the contents in a temporary folder,
    applies override contents, and then re-packages the contents into the dest
    file supplied as a parameter.

    :param source: A PyCdlib class object of an opened ISO file to extract.
    :param dest: The destination filename to write the result to.
    :param override_files: A dictionary of file name keys and contents stored
                           as values paris, which will be used to override
                           contents of the configuration drive for the purpose
                           of patching metadata.
    :param node_uuid: The node UUID, for logging purposes.
    """
    with tempfile.TemporaryDirectory(dir=CONF.tempdir) as tempfolder:
        # NOTE(TheJulia): The base extraction logic was sourced from kickstart
        # utils, as an actually kind of simple base approach to do the needful
        # extraction.
        LOG.debug('Extracting configuration drive for node %(node)s to '
                  '%(location)s.',
                  {'node': node_uuid,
                   'location': tempfolder})
        # The default configuration drive format we expect is generally
        # of a rock ridge format, and to avoid classic dos ISO9660 level
        # constraints, we will get the path using a rock ridge base path
        # so we don't have dashes translated into underscores on files
        # we want to preserve.
        for path, dirlist, filelist in source.walk(rr_path='/'):
            unix_path = path.lstrip('/')
            if path != "/":
                os.makedirs(os.path.join(tempfolder, unix_path), exist_ok=True)
            for f in filelist:
                # In iso9660 file extensions are mangled. Example '/FOO/BAR;1'.
                iso_file_path = os.path.join(path, f)
                file_record = source.get_record(rr_path=iso_file_path)
                posix_file_path = source.full_path_from_dirrecord(
                    file_record, rockridge=True
                )
                # Path to which the file in config drive to be written on the
                # server.
                posix_file_path = os.path.join(tempfolder,
                                               posix_file_path.lstrip('/'))
                # Extract the actual file from the ISO.
                source.get_file_from_iso(
                    joliet_path=iso_file_path, local_path=posix_file_path)
        # Okay, we have the files in a folder, lets patch!
        for file in override_files:
            LOG.debug('Patching override configuration drive file '
                      '%(file)s for node %(node)s.',
                      {'file': file,
                       'node': node_uuid})
            content = override_files[file]
            dest_path = os.path.join(tempfolder, str(file).lstrip('/'))
            with open(dest_path, mode="w") as new_file:
                new_file.write(content)
        # Instead of trying to use pycdlib to walk and assemble everything,
        # we'll just call out to mkisofs
        processutils.execute('mkisofs',
                             '-o', dest,
                             '-ldots',
                             '-allow-lowercase',
                             '-allow-multidot',
                             '-l',
                             '-publisher', 'Ironic',
                             '-quiet', '-J', '-r',
                             '-V', 'config-2',
                             tempfolder, attempts=1)


def generate_instance_network_data(task):
    """Generates OpenStack instance network metadata.

    This method was added to help facilitate the correction of cases
    where bad metadata could be supplied to Ironic by a service such
    as Nova where an instance has been requested, but missing
    network configuration details harms the viability of the overall
    instance's deployment.

    This method works by taking the available information which Ironic
    manages and controls, and reconciles it together into an OpenStack
    style network_data document which can then be rendered as
    network_data.json.

    :param task: A TaskManager class instance.
    :returns: A dictionary object with three keys, 'links', 'networks',
              and 'services' which contains various related entries
              as generated by the internal neutron get_neutron_port_data
              method which is then combined such that bond ports are also
              honored.
    """
    ports = task.ports
    portgroups = task.portgroups

    # get tenant vifs from all port objects, and reverse the structural
    # mapping so we keep the data associated for re-assembly.
    vif_id_to_objects = {'ports': {}, 'portgroups': {}}
    for name, collection in (('ports', ports), ('portgroups', portgroups)):
        for p in collection:
            # NOTE(TheJulia): While super similar to nova metadata code
            # as it pertiains to ironic, this explicitly focuses on
            vif_id = p.internal_info.get('tenant_vif_port_id')
            if vif_id:
                vif_id_to_objects[name][vif_id] = p
                # So now we have something like
                # vif_id_to_objects['ports'][vif_id] holding port object
                # so we have something easy and well structured to work with
                # through the rest of the method.
        # Where things are weird for nova and more clear-cut here is
        # we have and own all the physical structural data about the ports.
        # In nova, it is based upon what the user requests from an instance
        # and attempts to be bridge into the context of a baremetal node.
        # Metalsmith's metadata generation is geared similar to a nova
        # instance, the union of requested. Here, we don't really need to
        # take any guesses, everything is already mapped.

        # The pattern here is keyed off a list of VIF Ids. We'll need
        # this to finish hydrating necessary data.
        vif_id_keys = list(vif_id_to_objects['ports'])
        vif_id_keys.extend(list(vif_id_to_objects['portgroups']))

        # This is the starting point for metadata generation. We have three
        # buckets which contain lists of dicts. In this specific case, we
        # have to create a consolidated view across all attachments which
        # will require us to add values in.
    net_meta = {
        'links': [],
        'networks': [],
        'services': [],
    }
    for vif_id in vif_id_keys:
        vif_net_meta = {}
        if vif_id in vif_id_to_objects['portgroups']:
            pg = vif_id_to_objects['portgroups'][vif_id]
            pg_ports = [p for p in ports if p.portgroup_id == pg.id]
            bond_links = []
            # Pre-bake the various links for the bond.
            # This won't cause any duplicates to be added. A port
            # cannot be in more than one port group for the same
            # node.
            for p in pg_ports:
                # Assemble all of the portgroup members to ensure
                # they are present in the resulting list of physical
                # links to consider, so it can be assembled together.
                bond_links.append(
                    {
                        'id': p.uuid,
                        'type': 'phy',
                        'ethernet_mac_address': p.address,
                    }
                )
            vif_net_meta = neutron.get_neutron_port_data(
                pg.id, vif_id, iface_type='bond',
                mac_address=pg.address, bond_links=bond_links)
        elif vif_id in vif_id_to_objects['ports']:
            p = vif_id_to_objects['ports'][vif_id]
            vif_net_meta = neutron.get_neutron_port_data(
                p.id, vif_id,
                mac_address=p.address,
                iface_type='phy')
        else:
            LOG.error('While attempting to generate network metadata, '
                      'an unexpected case was encountered where VIF '
                      '%(vif)s was expected but then not found for '
                      'the deploy of node %(node)s.',
                      {'node': task.node.uuid,
                       'vif': vif_id})
        # Assemble the *per* VIF parts into the whole
        for key in ['links', 'networks', 'services']:
            # There is a fundimental issue, which is if
            # a *user* requests instances with multiple distinct
            # VIFs, and those vifs have multiple networks and
            # differences, it might not reconcile or may create
            # a conflicting situation. This is a reality and
            # there is likely not much we can do to  head it off
            # as its entirely a possible situation in an entirely
            # manually configured universe.
            net_meta[key].extend(vif_net_meta[key])

    # NOTE(TheJulia): This is done after the fact since our entire
    # structure assembles everything together by vif and upon the basis
    # of the vif attachments. We swap the values down to shorter values
    # because it turns out some configdrive readers will try to use the
    # id as a name value instead of operating relatively within the
    # context of the running state of the node where the data applies.
    link_count = 0
    link_map = {}
    # This is all modeled to modify in place so we minimally touch
    # the rest of the data structure.
    for link in net_meta['links']:
        link_name = 'iface{}'.format(link_count)
        link_map[link['id']] = link_name
        net_meta['links'][link_count]['id'] = link_name
        link_count += 1
    # We have to re-iterate through the all of the links to address
    # bond links, since we should already have mapped names, and now
    # we just need to correct them.
    link_count = 0
    for link in net_meta['links']:
        bond_links = link.get('bond_links')
        if bond_links:
            new_links = [link_map[x] for x in bond_links]
            net_meta['links'][link_count]['bond_links'] = new_links
        link_count += 0
    # We now need to re-associate networks to the new interface
    # id values.
    net_count = 0
    for network in net_meta['networks']:
        new_name = link_map[net_meta['networks'][net_count]['link']]
        net_meta['networks'][net_count]['link'] = new_name
        net_count += 1
    return net_meta


def generate_config_metadata(task):
    """Generates new network config metadata.

    This method wraps generate_instance_network_data and performs
    the logging actions in relation to it, and also performs a post
    generation sanity check on the network metadata.

    When the data is invalid, this method *also* raises an internal
    ConfigDriveRegenerationFailure exception which is a signal to the
    caller method that an issue has occurred. The expectation being
    that the caller knows how to handle and navigate that issue without
    failing the deployment operation.

    :param task: A TaskManager object.
    :returns: A dictionary object with three keys, 'links', 'networks',
              and 'services' which contains various related entries
              as generated by the internal neutron get_neutron_port_data
              method which is then combined such that bond ports are also
              honored.
    :raises: ConfigDriveRegenerationFailure when no metadata is generated
             or the configuration state of the drivers does not support
             the generation of network metadata. The purpose of this is
             to appropriately signal an error to the caller, not fail the
             overall requested operation by the end user.
    """
    # NOTE(TheJulia): This method could be improved or extended in the
    # future, and with that somewhat in mind, the name of the method
    # was left appropriately broad to set that stage.

    metadata = generate_instance_network_data(task)
    if is_invalid_network_metadata(metadata):
        # NOTE(TheJulia): This *should* never happen in the scenarios
        # known to Ironic, however *could* happen if there is some sort
        # of weird state due to drivers. i.e. if someone had a completely
        # custom set of interfaces, this could occur and we don't want
        # this to be fatal because maybe this is also the intent. In
        # any event, log the error and move on.
        LOG.error('Failed to generate new network metadata for '
                  'deployment of node %s. No virtual interfaces '
                  'were generated. Possible missing vifs?',
                  task.node.uuid)
        raise exception.ConfigDriveRegenerationFailure()
    return metadata


def check_and_fix_configdrive(task, configdrive):
    """Check and fix the supplied config drive.

    In certain cases, a caller may supply a horribly broken configuration
    drive which contains a list of empty dicts or even is an ISO in the same
    state.

    The prupose of this method is to attempt to identify and correct those
    specific cases to help ensure the instance being deployed has what is
    necessary to boot.

    If this method fails, an internal exception,
    ConfigDriveRegenerationFailure is raised by the various helper methods
    used by this method, however this method will just return the user
    supplied

    :param task: A Taskmanager object.
    :param configdrive: The supplied configuration drive.
    :returns: Returns a configuration drive aligning with the type which
              was originally supplied by the user.
    """

    # NOTE(TheJulia): A possible, feature sort of expansion is to potentially
    # use this same method to also just supply network metadata injection if
    # no configuration drive is supplied.
    try:
        if isinstance(configdrive, str) and configdrive.startswith('http'):
            # In this event, we've been given a URL. We don't support
            # trying to do anything else hwere.
            return configdrive
        elif isinstance(configdrive, dict):
            provided_network_data = configdrive.get('network_data')
            # TODO(TheJulia): This is where we could inject network config
            # metadata into config drive results if not supplied to by the
            # user.
            if is_invalid_network_metadata(provided_network_data):
                # BUG 2106073 OR there is data, but no network data in the
                # configuration drive. In such a case, we need to take what
                # we know to be the metadata based upon the state,
                # and replace the field.
                generated_metadata = generate_config_metadata(task)
                # Explicitly set to a variable to allow the method to not
                # possibly impact the original configdrive value.
                configdrive['network_data'] = generated_metadata
        else:
            # This method knows how to patch the existing configdrive and
            # call the metadata check independently due to the complexity
            # of handling the config drive ISOs and fixing them when they
            # need updates.
            LOG.debug('Starting to evaluate configdrive for node %s',
                      task.node.uuid)
            configdrive = check_and_patch_configdrive(task, configdrive)
    except exception.ConfigDriveRegenerationFailure:
        LOG.warning('Ironic was unable to regenerate the configdrive for node '
                    '%s.', task.node.uuid)
        # This is not a fatal failure, just circle things back.
    # Always return configdrive content, otherwise we hand None back
    # which causes the overall process to fail.
    return configdrive



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Redfish Inspect Interface
"""

from oslo_log import log
from oslo_utils import units
import sushy

from ironic.common import boot_modes
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import engine
from ironic.common import states
from ironic.common import utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.drivers import utils as drivers_utils

LOG = log.getLogger(__name__)

PROCESSOR_INSTRUCTION_SET_MAP = {
    sushy.InstructionSet.ARM_A32: 'arm',
    sushy.InstructionSet.ARM_A64: 'aarch64',
    sushy.InstructionSet.IA_64: 'ia64',
    sushy.InstructionSet.MIPS32: 'mips',
    sushy.InstructionSet.MIPS64: 'mips64',
    sushy.InstructionSet.OEM: None,
    sushy.InstructionSet.X86: 'i686',
    sushy.InstructionSet.X86_64: 'x86_64'
}

BOOT_MODE_MAP = {
    sushy.BOOT_SOURCE_MODE_UEFI: boot_modes.UEFI,
    sushy.BOOT_SOURCE_MODE_BIOS: boot_modes.LEGACY_BIOS
}


class RedfishInspect(base.InspectInterface):

    def __init__(self):
        super().__init__()
        enabled_hooks = [x.strip()
                         for x in CONF.redfish.inspection_hooks.split(',')
                         if x.strip()]
        self.hooks = inspect_utils.validate_inspection_hooks("redfish",
                                                             enabled_hooks)

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validate the driver-specific Node deployment info.

        This method validates whether the 'driver_info' properties of
        the task's node contains the required information for this
        interface to function.

        This method is often executed synchronously in API requests, so it
        should not conduct long-running checks.

        :param task: A TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        redfish_utils.parse_driver_info(task.node)

    def inspect_hardware(self, task):
        """Inspect hardware to get the hardware properties.

        Inspects hardware to get the essential properties.
        It fails if any of the essential properties
        are not received from the node.

        :param task: a TaskManager instance.
        :raises: HardwareInspectionFailure if essential properties
                 could not be retrieved successfully.
        :returns: The resulting state of inspection.

        """
        system = redfish_utils.get_system(task.node)

        # get the essential properties and update the node properties
        # with it.
        inspected_properties = task.node.properties
        inventory = {}

        if system.memory_summary and system.memory_summary.size_gib:
            memory = system.memory_summary.size_gib * units.Ki
            inspected_properties['memory_mb'] = memory
            inventory['memory'] = {'physical_mb': memory}

        # match the inventory data of ironic-inspector / ironic-python-agent
        # to make existing inspection hooks and rules work by defaulting
        # the values
        inventory['cpu'] = {
            'count': 0,
            'architecture': '',
        }
        proc_info = self._get_processor_info(task, system)
        inventory['cpu'].update(proc_info)

        # TODO(etingof): should we respect root device hints here?
        local_gb = self._detect_local_gb(task, system)

        if local_gb:
            inspected_properties['local_gb'] = str(local_gb)

        else:
            LOG.warning("Could not provide a valid storage size configured "
                        "for node %(node)s. Assuming this is a disk-less node",
                        {'node': task.node.uuid})
            inspected_properties['local_gb'] = '0'

        if storages := system.storage or system.simple_storage:
            disks = list()
            for storage in storages.get_members():
                drives = storage.drives if hasattr(
                    storage, 'drives') else storage.devices
                for drive in drives:
                    disk = {}
                    disk['name'] = drive.name
                    disk['size'] = drive.capacity_bytes
                    disks.append(disk)

            inventory['disks'] = disks

        if system.ethernet_interfaces and system.ethernet_interfaces.summary:
            inventory['interfaces'] = []
            for eth in system.ethernet_interfaces.get_members():
                iface = {'mac_address': eth.mac_address,
                         'name': eth.identity}
                inventory['interfaces'].append(iface)

        system_vendor = {}
        if system.name:
            system_vendor['product_name'] = str(system.name)

        if system.serial_number:
            system_vendor['serial_number'] = str(system.serial_number)

        if system.manufacturer:
            system_vendor['manufacturer'] = str(system.manufacturer)

        if system_vendor:
            inventory['system_vendor'] = system_vendor

        if system.boot.mode:
            if not drivers_utils.get_node_capability(task.node, 'boot_mode'):
                capabilities = utils.get_updated_capabilities(
                    inspected_properties.get('capabilities', ''),
                    {'boot_mode': BOOT_MODE_MAP[system.boot.mode]})

                inspected_properties['capabilities'] = capabilities
            inventory['boot'] = {'current_boot_mode':
                                 BOOT_MODE_MAP[system.boot.mode]}

        self._create_ports(task, system)

        pxe_port_macs = self._get_pxe_port_macs(task)
        # existing data format only allows one mac so use that for now
        if pxe_port_macs:
            inventory['boot']['pxe_interface'] = pxe_port_macs[0]

        plugin_data = {}

        inspect_utils.run_inspection_hooks(task, inventory, plugin_data,
                                           self.hooks, None)
        inspect_utils.store_inspection_data(task.node,
                                            inventory,
                                            plugin_data,
                                            task.context)
        engine.apply_rules(task, inventory, plugin_data, 'main')

        valid_keys = self.ESSENTIAL_PROPERTIES
        missing_keys = valid_keys - set(inspected_properties)
        if missing_keys:
            error = (_('Failed to discover the following properties: '
                       '%(missing_keys)s on node %(node)s') %
                     {'missing_keys': ', '.join(missing_keys),
                      'node': task.node.uuid})
            raise exception.HardwareInspectionFailure(error=error)

        task.node.properties = inspected_properties
        task.node.save()
        LOG.debug("Node properties for %(node)s are updated as "
                  "%(properties)s", {'properties': inspected_properties,
                                     'node': task.node.uuid})

        return states.MANAGEABLE

    def _create_ports(self, task, system):
        enabled_macs = redfish_utils.get_enabled_macs(task, system)
        if enabled_macs:
            inspect_utils.create_ports_if_not_exist(task, list(enabled_macs))
        else:
            LOG.warning("Not attempting to create any port as no NICs "
                        "were discovered in 'enabled' state for node "
                        "%(node)s: %(mac_data)s",
                        {'mac_data': enabled_macs, 'node': task.node.uuid})

    def _detect_local_gb(self, task, system):
        simple_storage_size = 0

        try:
            LOG.debug("Attempting to discover system simple storage size for "
                      "node %(node)s", {'node': task.node.uuid})
            if (system.simple_storage
                    and system.simple_storage.disks_sizes_bytes):
                simple_storage_size = [
                    size for size in system.simple_storage.disks_sizes_bytes
                    if size >= 4 * units.Gi
                ] or [0]

                simple_storage_size = simple_storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No simple storage information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        storage_size = 0

        try:
            LOG.debug("Attempting to discover system storage volume size for "
                      "node %(node)s", {'node': task.node.uuid})
            if system.storage and system.storage.volumes_sizes_bytes:
                storage_size = [
                    size for size in system.storage.volumes_sizes_bytes
                    if size >= 4 * units.Gi
                ] or [0]

                storage_size = storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No storage volume information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        try:
            if not storage_size:
                LOG.debug("Attempting to discover system storage drive size "
                          "for node %(node)s", {'node': task.node.uuid})
                if system.storage and system.storage.drives_sizes_bytes:
                    storage_size = [
                        size for size in system.storage.drives_sizes_bytes
                        if size >= 4 * units.Gi
                    ] or [0]

                    storage_size = storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No storage drive information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        # NOTE(etingof): pick the smallest disk larger than 4G among available
        if simple_storage_size and storage_size:
            local_gb = min(simple_storage_size, storage_size)

        else:
            local_gb = max(simple_storage_size, storage_size)

        # Note(deray): Convert the received size to GiB and reduce the
        # value by 1 GB as consumers like Ironic requires the ``local_gb``
        # to be returned 1 less than actual size.
        return max(0, int(local_gb / units.Gi - 1))

    def _get_pxe_port_macs(self, task):
        """Get a list of PXE port MAC addresses.

        :param task: a TaskManager instance.
        :returns: Returns list of PXE port MAC addresses.
                  If cannot be determined, returns None.
        """
        return None

    def _get_processor_info(self, task, system):
        # NOTE(JayF): Checking truthiness here is better than checking for None
        #             because if we have an empty list, we'll raise a
        #             ValueError.
        cpu = {}

        if not system.processors:
            return cpu

        if system.processors.summary:
            cpu['count'], _ = system.processors.summary

        processor = system.processors.get_members()[0]

        if processor.model is not None:
            cpu['model_name'] = str(processor.model)
        if processor.max_speed_mhz is not None:
            cpu['frequency'] = processor.max_speed_mhz
        cpu['architecture'] = PROCESSOR_INSTRUCTION_SET_MAP.get(
            processor.instruction_set) or ''

        return cpu



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import gzip
import json
import os
import tempfile
from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import uuidutils
import pycdlib

from ironic.common import exception
from ironic.common import neutron
from ironic.conductor import configdrive_utils as cd_utils
from ironic.conductor import task_manager
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class MetadataUtilsTestCase(db_base.DbTestCase):

    def setUp(self):
        super(MetadataUtilsTestCase, self).setUp()
        self.node = obj_utils.create_test_node(
            self.context, driver='fake-hardware',
            uuid=uuidutils.generate_uuid())

    def test_is_invalid_network_metadata(self):
        """Test the known invalid state."""
        invalid = {'links': [], 'services': [], 'networks': []}
        self.assertTrue(cd_utils.is_invalid_network_metadata(invalid))

    def test_invalid_network_metadata_list(self):
        self.assertTrue(cd_utils.is_invalid_network_metadata([]))

    def test_is_invalid_network_metadata_valid(self):
        valid = {'links': [{'foo': 'bar', 'mtu': 1500}],
                 'services': [],
                 'networks': [{'bar': 'baz'}]}
        self.assertFalse(cd_utils.is_invalid_network_metadata(valid))

    def test_invalid_network_metadata_null_mtu(self):
        invalid = {'links': [{'mtu': None}],
                   'servies': [],
                   'networks': [{'foo': 'bar'}]}
        self.assertTrue(cd_utils.is_invalid_network_metadata(invalid))

    def test_invalid_network_metadata_null_mtu_disables(self):
        CONF.set_override('disable_metadata_mtu_check', True,
                          group='conductor')
        invalid = {'links': [{'mtu': None}],
                   'servies': [],
                   'networks': [{'foo': 'bar'}]}
        self.assertFalse(cd_utils.is_invalid_network_metadata(invalid))

    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_instance_network_data',
                       autospec=True)
    def test_generate_config_metadata(self, mock_gen, mock_valid):
        # Just enough to be something
        fake_data = {'foo': 'bar'}
        mock_gen.return_value = fake_data
        mock_valid.return_value = False
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            res = cd_utils.generate_config_metadata(task)
            mock_gen.assert_called_once_with(task)
        self.assertEqual(fake_data, res)
        mock_valid.assert_called_once_with(fake_data)

    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_instance_network_data',
                       autospec=True)
    def test_generate_config_metadata_none(self, mock_gen, mock_valid):
        # We don't expect None to ever be returned from
        # generate_instance_network_data, but just in case!
        mock_gen.return_value = None
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            self.assertRaises(
                exception.ConfigDriveRegenerationFailure,
                cd_utils.generate_config_metadata,
                task)
            mock_gen.assert_called_once_with(task)
        mock_valid.assert_called_once_with(None)

    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_instance_network_data',
                       autospec=True)
    def test_generate_config_metadata_invalid(self, mock_gen, mock_valid):
        fake_data = {'foo': 'bar'}
        mock_gen.return_value = fake_data
        mock_valid.return_value = True
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            self.assertRaises(
                exception.ConfigDriveRegenerationFailure,
                cd_utils.generate_config_metadata,
                task)
            mock_gen.assert_called_once_with(task)
        mock_valid.assert_called_once_with(fake_data)

    def test_generate_instance_network_data(self):
        # This is just a test which validates we pass through the method
        # as we expect with no vifs.
        self.node.network_interface = 'noop'
        self.node.save()
        expected_dict = {'links': [], 'networks': [], 'services': []}
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            self.assertNotIn('metadata', task.driver.network.capabilities)
            self.assertDictEqual(
                expected_dict,
                cd_utils.generate_instance_network_data(task))

    def test_generate_instance_network_data_no_vif(self):
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            self.assertEqual(
                {'links': [], 'networks': [], 'services': []},
                cd_utils.generate_instance_network_data(task))

    @mock.patch.object(neutron, 'get_neutron_port_data', autospec=True)
    def test_generate_instance_network_data_single_vif(self, mock_gnpd):
        port = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='86ffb399-6dd4-4e6c-8eb0-069d608bc5ca',
            internal_info={'tenant_vif_port_id': 'meep'})

        gnpd = {
            'links': [
                {
                    'id': port.uuid,
                    'type': 'vif',
                    'ethernet_mac_address': '01:02:03:04:05:06',
                    'vif_id': 'meep',
                    'mtu': 1500
                }
            ],
            'networks': [
                {
                    'id': "boop",
                    'network_id': "a8164a5e-ce7e-4ce4-b017-20c93a559f7c",
                    'type': 'ipv4',
                    'link': port.uuid,
                    'ip_address': "192.168.1.1",
                    'netmask': "255.255.255.0",
                    'routes': []
                }
            ],
            'services': []
        }
        mock_gnpd.return_value = gnpd
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            res = cd_utils.generate_instance_network_data(task)
            # NOTE(TheJulia): Normally, we would expect the method to
            # return slightly different data, specifically superseded
            # mac address and a phy, but the test here is the consolidation
            # of result data together.
            self.assertEqual(gnpd, res)
            mock_gnpd.assert_called_once_with(port.id, 'meep',
                                              mac_address='52:54:00:cf:2d:31',
                                              iface_type='phy')

    @mock.patch.object(neutron, 'get_neutron_port_data', autospec=True)
    def test_generate_instance_network_data_multi_vif(self, mock_gnpd):
        port1 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='86ffb399-6dd4-4e6c-8eb0-069d608bc5ca',
            internal_info={'tenant_vif_port_id': 'meep'})

        port2 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='373960ab-131f-4a37-8329-8cda768ba722',
            internal_info={'tenant_vif_port_id': 'beep'},
            address="51:54:00:cf:2d:32")

        gnpd1 = {
            'links': [
                {'id': port1.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': port1.address,
                 'vif_id': 'meep',
                 'mtu': 1500}],
            'networks': [
                {'id': "boop",
                 'network_id': "a8164a5e-ce7e-4ce4-b017-20c93a559f7c",
                 'type': 'ipv4',
                 'link': port1.uuid,
                 'ip_address': "192.168.1.1",
                 'netmask': "255.255.255.0",
                 'routes': []}],
            'services': []
        }
        gnpd2 = {
            'links': [
                {'id': port2.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': port2.address,
                 'vif_id': 'beep',
                 'mtu': 1500}],
            'networks': [
                {'id': "boop",
                 'network_id': "a8164a5e-ce7e-4ce4-b017-20c93a559f7c",
                 'type': 'ipv4',
                 'link': port2.uuid,
                 'ip_address': "192.168.1.2",
                 'netmask': "255.255.255.0",
                 'routes': []}],
            'services': [{"address": "8.8.8.8", "type": "dns"}]
        }

        expected = {
            'links': [
                {'ethernet_mac_address': '52:54:00:cf:2d:31',
                 'id': 'iface0',
                 'mtu': 1500,
                 'type': 'phy',
                 'vif_id': 'meep'},
                {'ethernet_mac_address': '51:54:00:cf:2d:32',
                 'id': 'iface1',
                 'mtu': 1500,
                 'type': 'phy',
                 'vif_id': 'beep'}],
            'networks': [
                {'id': 'boop',
                 'ip_address': '192.168.1.1',
                 'link': 'iface0',
                 'netmask': '255.255.255.0',
                 'network_id': 'a8164a5e-ce7e-4ce4-b017-20c93a559f7c',
                 'routes': [],
                 'type': 'ipv4'},
                {'id': 'boop',
                 'ip_address': '192.168.1.2',
                 'link': 'iface1',
                 'netmask': '255.255.255.0',
                 'network_id': 'a8164a5e-ce7e-4ce4-b017-20c93a559f7c',
                 'routes': [],
                 'type': 'ipv4'}],
            'services': [{'address': '8.8.8.8', 'type': 'dns'}]
        }

        mock_gnpd.side_effect = [gnpd1, gnpd2]
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            res = cd_utils.generate_instance_network_data(task)
            # NOTE(TheJulia): Normally, we would expect the method to
            # return slightly different data, specifically superseded
            # mac address and a phy, but the test here is the consolidation
            # of result data together.
            self.assertEqual(expected, res)
            mock_gnpd.assert_has_calls([
                mock.call(port1.id, 'meep',
                          mac_address=port1.address,
                          iface_type='phy'),
                mock.call(port2.id, 'beep',
                          mac_address=port2.address,
                          iface_type='phy')])

    @mock.patch.object(neutron, 'get_neutron_port_data', autospec=True)
    def test_generate_instance_network_data_portgroup(self, mock_gnpd):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            uuid='f52c664d-ca02-4c27-94f9-327dcb67d208',
            address='01:02:03:04:05:06',
            internal_info={'tenant_vif_port_id': 'meep'})

        port1 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='86ffb399-6dd4-4e6c-8eb0-069d608bc5ca',
            portgroup_id=pg.id)

        port2 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='373960ab-131f-4a37-8329-8cda768ba722',
            address="51:54:00:cf:2d:32",
            portgroup_id=pg.id)

        gnpd = {
            'links': [
                {'id': pg.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': '01:02:03:04:05:06',
                 'vif_id': 'meep',
                 'mtu': 1500,
                 'bond_links': [port1.uuid, port2.uuid]},
                {'id': port1.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': port1.address,
                 'mtu': 1500},
                {'id': port2.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': port2.address,
                 'mtu': 1500}],
            'networks': [
                {'id': "boop",
                 'network_id': "a8164a5e-ce7e-4ce4-b017-20c93a559f7c",
                 'type': 'ipv4',
                 'link': pg.uuid,
                 'ip_address': "192.168.1.1",
                 'netmask': "255.255.255.0",
                 'routes': []}],
            'services': []
        }
        expected = {
            'links': [
                {'bond_links': ['iface1', 'iface2'],
                 'ethernet_mac_address': '01:02:03:04:05:06',
                 'id': 'iface0',
                 'mtu': 1500,
                 'type': 'phy',
                 'vif_id': 'meep'},
                {'ethernet_mac_address': '52:54:00:cf:2d:31',
                 'id': 'iface1',
                 'mtu': 1500,
                 'type': 'phy'},
                {'ethernet_mac_address': '51:54:00:cf:2d:32',
                 'id': 'iface2',
                 'mtu': 1500,
                 'type': 'phy'}],
            'networks': [{
                'id': 'boop',
                'ip_address': '192.168.1.1',
                'link': 'iface0',
                'netmask': '255.255.255.0',
                'network_id': 'a8164a5e-ce7e-4ce4-b017-20c93a559f7c',
                'routes': [],
                'type': 'ipv4'}],
            'services': []}

        mock_gnpd.return_value = gnpd
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            res = cd_utils.generate_instance_network_data(task)
            self.assertEqual(expected, res)
            mock_gnpd.assert_called_once_with(
                pg.id, 'meep',
                mac_address=pg.address,
                iface_type='bond',
                bond_links=[{'id': port1.uuid, 'type': 'phy',
                             'ethernet_mac_address': port1.address},
                            {'id': port2.uuid, 'type': 'phy',
                             'ethernet_mac_address': port2.address}])

    @mock.patch.object(neutron, 'get_neutron_port_data', autospec=True)
    def test_generate_instance_network_data_portgroups(self, mock_gnpd):
        pg1 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            uuid='f52c664d-ca02-4c27-94f9-327dcb67d208',
            address='01:02:03:04:05:06',
            internal_info={'tenant_vif_port_id': 'meep'})
        port1 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='86ffb399-6dd4-4e6c-8eb0-069d608bc5ca',
            portgroup_id=pg1.id)
        pg2 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            name='second', id=2,
            uuid='c1ef2f0b-36f8-45f6-9b0c-c68bd27beee1',
            address='00:02:03:04:05:07',
            internal_info={'tenant_vif_port_id': 'flop'})
        port2 = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid='373960ab-131f-4a37-8329-8cda768ba722',
            address="51:54:00:cf:2d:32",
            portgroup_id=pg2.id)

        gnpd = {
            'links': [
                {'id': pg1.uuid,
                 'type': 'phy',
                 'ethernet_mac_address': '01:02:03:04:05:06',
                 'vif_id': 'meep',
                 'mtu': 1500}],
            'networks': [
                {'id': "boop",
                 'network_id': "a8164a5e-ce7e-4ce4-b017-20c93a559f7c",
                 'type': 'ipv4',
                 'link': pg1.uuid,
                 'ip_address': "192.168.1.1",
                 'netmask': "255.255.255.0",
                 'routes': []}],
            'services': []
        }
        # Return something so the rest of the code loops properly
        mock_gnpd.return_value = gnpd
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            cd_utils.generate_instance_network_data(task)
            # NOTE(TheJulia): Not really concerned with the result, the
            # key aspect is to ensure we assembled the bond links correctly
            # from the available data. Other tests ensure the fields
            # get assembled together.
            mock_gnpd.assert_has_calls([
                mock.call(pg1.id, 'meep',
                          mac_address=pg1.address,
                          iface_type='bond',
                          bond_links=[
                              {'id': port1.uuid, 'type': 'phy',
                               'ethernet_mac_address': port1.address}]),
                mock.call(pg2.id, 'flop',
                          mac_address=pg2.address,
                          iface_type='bond',
                          bond_links=[
                              {'id': port2.uuid, 'type': 'phy',
                               'ethernet_mac_address': port2.address}])])

    @mock.patch.object(cd_utils, '_read_config_drive', autospec=True)
    @mock.patch.object(cd_utils, 'regenerate_iso',
                       autospec=True)
    @mock.patch.object(base64, 'b64decode', autospec=True)
    @mock.patch.object(gzip, 'decompress', autospec=True)
    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_config_metadata',
                       autospec=True)
    @mock.patch.object(tempfile, 'mkstemp', autospec=True)
    @mock.patch.object(tempfile, 'NamedTemporaryFile', autospec=True)
    @mock.patch.object(os, 'remove', autospec=True)
    @mock.patch.object(pycdlib, 'PyCdlib', autospec=True)
    def test_check_and_patch_configdrive(
            self,
            mock_pycd, mock_remove,
            mock_temp, mock_mkstemp,
            mock_gen, mock_is_invalid,
            mock_decomp,
            mock_b64_decode,
            mock_regen,
            mock_read_iso):
        invalid_nd = json.dumps({'links': [], 'services': [], 'networks': []})
        mock_pycd.return_value.open_file_from_iso.return_value.__enter__. \
            return_value.read.return_value = invalid_nd
        mock_is_invalid.return_value = True
        mock_gen.return_value = {'foo': 'bar'}
        mock_regen.return_value = '{"foo": "bar"}'
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            cd_utils.check_and_patch_configdrive(task, 'foo')
        mock_regen.assert_called_once_with(
            mock.ANY, mock.ANY,
            {'/openstack/latest/network_data.json': mock.ANY},
            node_uuid=self.node.uuid)
        self.assertTrue(mock_remove.called)
        mock_temp.assert_called_once_with(dir=mock.ANY, mode='wb+')
        mock_mkstemp.assert_called_once_with(dir=mock.ANY)
        self.assertTrue(mock_b64_decode.called)
        self.assertTrue(mock_decomp.called)
        self.assertTrue(mock_is_invalid.called)
        mock_read_iso.assert_called_once_with(mock.ANY)

    @mock.patch.object(cd_utils, '_read_config_drive', autospec=True)
    @mock.patch.object(cd_utils, 'regenerate_iso',
                       autospec=True)
    @mock.patch.object(base64, 'b64decode', autospec=True)
    @mock.patch.object(gzip, 'decompress', autospec=True)
    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_config_metadata',
                       autospec=True)
    @mock.patch.object(tempfile, 'mkstemp', autospec=True)
    @mock.patch.object(tempfile, 'NamedTemporaryFile', autospec=True)
    @mock.patch.object(os, 'remove', autospec=True)
    @mock.patch.object(pycdlib, 'PyCdlib', autospec=True)
    def test_check_and_patch_configdrive_invalid_nework_data(
            self,
            mock_pycd, mock_remove,
            mock_temp, mock_mkstemp,
            mock_gen, mock_is_valid,
            mock_decomp,
            mock_b64_decode,
            mock_regen,
            mock_read_iso):
        invalid_nd = '{"foo":...'
        mock_pycd.return_value.open_file_from_iso.return_value.__enter__. \
            return_value.read.return_value = invalid_nd
        mock_gen.return_value = {'foo': 'bar'}
        mock_regen.return_value = '{"foo": "bar"}'
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            cd_utils.check_and_patch_configdrive(task, 'foo')
        mock_regen.assert_called_once_with(
            mock.ANY, mock.ANY,
            {'/openstack/latest/network_data.json': mock.ANY},
            node_uuid=self.node.uuid)
        self.assertTrue(mock_remove.called)
        mock_temp.assert_called_once_with(dir=mock.ANY, mode='wb+')
        mock_mkstemp.assert_called_once_with(dir=mock.ANY)
        self.assertTrue(mock_b64_decode.called)
        self.assertTrue(mock_decomp.called)
        self.assertFalse(mock_is_valid.called)
        mock_read_iso.assert_called_once_with(mock.ANY)

    @mock.patch.object(cd_utils, '_read_config_drive', autospec=True)
    @mock.patch.object(cd_utils, 'regenerate_iso',
                       autospec=True)
    @mock.patch.object(base64, 'b64decode', autospec=True)
    @mock.patch.object(gzip, 'decompress', autospec=True)
    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_config_metadata',
                       autospec=True)
    @mock.patch.object(tempfile, 'mkstemp', autospec=True)
    @mock.patch.object(tempfile, 'NamedTemporaryFile', autospec=True)
    @mock.patch.object(os, 'remove', autospec=True)
    @mock.patch.object(pycdlib, 'PyCdlib', autospec=True)
    def test_check_and_patch_configdrive_pycdlib_failure(
            self,
            mock_pycd, mock_remove,
            mock_temp, mock_mkstemp,
            mock_gen, mock_is_valid,
            mock_decomp,
            mock_b64_decode,
            mock_regen,
            mock_read_iso):
        mock_pycd.side_effect = \
            pycdlib.pycdlibexception.PyCdlibInvalidInput(msg='failure')
        mock_gen.return_value = {'foo': 'bar'}
        mock_regen.return_value = '{"foo": "bar"}'
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            cd_utils.check_and_patch_configdrive(task, 'foo')
        mock_regen.assert_not_called()
        mock_remove.assert_not_called()
        mock_temp.assert_called_once_with(dir=mock.ANY, mode='wb+')
        mock_mkstemp.assert_not_called()
        self.assertTrue(mock_b64_decode.called)
        self.assertTrue(mock_decomp.called)
        mock_is_valid.assert_not_called()
        mock_read_iso.assert_not_called()

    @mock.patch.object(cd_utils, '_read_config_drive', autospec=True)
    @mock.patch.object(cd_utils, 'regenerate_iso',
                       autospec=True)
    @mock.patch.object(base64, 'b64decode', autospec=True)
    @mock.patch.object(gzip, 'decompress', autospec=True)
    @mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                       autospec=True)
    @mock.patch.object(cd_utils, 'generate_config_metadata',
                       autospec=True)
    @mock.patch.object(tempfile, 'mkstemp', autospec=True)
    @mock.patch.object(tempfile, 'NamedTemporaryFile', autospec=True)
    @mock.patch.object(os, 'remove', autospec=True)
    @mock.patch.object(pycdlib, 'PyCdlib', autospec=True)
    def test_check_and_patch_configdrive_metadata_gen_fail(
            self,
            mock_pycd, mock_remove,
            mock_temp, mock_mkstemp,
            mock_gen, mock_is_valid,
            mock_decomp,
            mock_b64_decode,
            mock_regen,
            mock_read_iso):
        invalid_nd = json.dumps({'links': [], 'services': [], 'networks': []})
        mock_pycd.return_value.open_file_from_iso.return_value.__enter__. \
            return_value.read.return_value = invalid_nd
        mock_is_valid.return_value = False
        mock_gen.side_effect = exception.ConfigDriveRegenerationFailure
        mock_regen.return_value = '{"foo": "bar"}'
        with task_manager.acquire(self.context, self.node.id,
                                  shared=True) as task:
            cd_utils.check_and_patch_configdrive(task, 'foo')
        mock_regen.assert_not_called()
        mock_remove.assert_not_called()
        mock_temp.assert_called_once_with(dir=mock.ANY, mode='wb+')
        mock_mkstemp.assert_not_called()
        self.assertTrue(mock_b64_decode.called)
        self.assertTrue(mock_decomp.called)
        self.assertTrue(mock_is_valid.called)
        mock_read_iso.assert_not_called()

    @mock.patch('builtins.open', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    @mock.patch.object(tempfile, 'TemporaryDirectory', autospec=True)
    @mock.patch.object(processutils, 'execute', autospec=True)
    def test_regenerate_iso(self, mock_exec, mock_tempdir, mock_mkdirs,
                            mock_open):
        fake_pycd = mock.Mock()
        fake_pycd.walk.return_value = [('/', ['path'], []),
                                       ('/path', [], ['file']),
                                       ('/path/path2', [], [])]
        fake_pycd.get_record.return_value = 'boop'
        fake_pycd.full_path_from_dirrecord.return_value = 'foo'
        mock_tempdir.return_value.__enter__.return_value = '/tmp/temp_folder'
        mock_write = mock.Mock()
        mock_open.return_value.__enter__.return_value.write = mock_write
        cd_utils.regenerate_iso(
            fake_pycd, '/tmp/foo_file', {'/path/foo.txt': 'meow'})
        mock_exec.assert_called_with(
            'mkisofs',
            '-o', '/tmp/foo_file',
            '-ldots',
            '-allow-lowercase',
            '-allow-multidot',
            '-l',
            '-publisher', 'Ironic',
            '-quiet', '-J', '-r',
            '-V', 'config-2',
            '/tmp/temp_folder', attempts=1)
        mock_mkdirs.assert_has_calls([
            mock.call('/tmp/temp_folder/path', exist_ok=True),
            mock.call('/tmp/temp_folder/path/path2', exist_ok=True)])
        mock_write.assert_has_calls([mock.call('meow')])
        mock_open.assert_called_once_with('/tmp/temp_folder/path/foo.txt',
                                          mode='w')
        fake_pycd.get_record.assert_has_calls([
            mock.call(rr_path='/path/file')])
        fake_pycd.full_path_from_dirrecord.assert_called_once_with(
            'boop', rockridge=True)


@mock.patch.object(cd_utils, 'is_invalid_network_metadata',
                   autospec=True)
@mock.patch.object(cd_utils, 'generate_config_metadata',
                   autospec=True)
@mock.patch.object(cd_utils, 'check_and_patch_configdrive',
                   autospec=True)
class PatchConfigDriveTestCase(db_base.DbTestCase):

    def setUp(self):
        super(PatchConfigDriveTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='fake-hardware',
                                               instance_info=None)
        self.bad_drive_dict = {
            'network_data': {
                'links': [],
                'networks': [],
                'services': []
            }
        }

        self.mock_network_data = {
            'links': [{'foo': 'bar'}],
            'networks': [{'bar': 'baz'}],
            'services': [],
        }

    def test_check_and_fix_configdrive_metadata(self, mock_cdp, mock_gen,
                                                mock_invalid):
        mock_invalid.return_value = True
        mock_gen.return_value = self.mock_network_data
        with task_manager.acquire(self.context, self.node.uuid) as task:
            res = cd_utils.check_and_fix_configdrive(
                task, self.bad_drive_dict)
            mock_gen.assert_called_once_with(task)
        mock_invalid.assert_called_once_with({
            'links': [], 'networks': [], 'services': []})
        mock_cdp.assert_not_called()
        self.assertEqual({'network_data': self.mock_network_data}, res)

    def test_check_and_fix_configdrive_string(self, mock_cdp, mock_gen,
                                              mock_invalid):
        mock_cdp.return_value = 'foo2'
        with task_manager.acquire(self.context, self.node.uuid) as task:
            res = cd_utils.check_and_fix_configdrive(task, 'foo')
            mock_cdp.assert_called_once_with(task, 'foo')
        mock_invalid.assert_not_called()
        mock_gen.assert_not_called()
        self.assertEqual('foo2', res)

    def test_check_and_fix_configdrive_string_url(
            self, mock_cdp, mock_gen,
            mock_invalid):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            res = cd_utils.check_and_fix_configdrive(
                task, 'https://bifrost/url')
            mock_cdp.assert_not_called()
        mock_invalid.assert_not_called()
        mock_gen.assert_not_called()
        self.assertEqual('https://bifrost/url', res)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_service import sslutils

from ironic.common import exception
from ironic.common import wsgi_service
from ironic.tests import base

CONF = cfg.CONF


class TestWSGIService(base.TestCase):
    def setUp(self):
        super().setUp()

        sslutils.register_opts(CONF)
        self.server = mock.Mock()
        self.server.requests = mock.Mock(min=0, max=0)

    @mock.patch.object(processutils, 'get_worker_count', lambda: 2)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_default(self, mock_server):
        service_name = "ironic_api"
        mock_server.return_value = self.server
        test_service = wsgi_service.WSGIService(service_name)
        self.assertEqual(2, test_service.workers)
        mock_server.assert_called_once_with(server_name=service_name,
                                            wsgi_app=test_service.app,
                                            bind_addr=('0.0.0.0', 6385))

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_correct_setting(self, mock_server):
        self.config(api_workers=8, group='api')
        mock_server.return_value = self.server
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(8, test_service.workers)

    @mock.patch.object(processutils, 'get_worker_count', lambda: 3)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_zero_setting(self, mock_server):
        self.config(api_workers=0, group='api')
        mock_server.return_value = self.server
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(3, test_service.workers)

    @mock.patch.object(processutils, 'get_worker_count', lambda: 42)
    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_default_limit(self, mock_server):
        self.config(api_workers=0, group='api')
        mock_server.return_value = self.server
        test_service = wsgi_service.WSGIService("ironic_api")
        self.assertEqual(4, test_service.workers)

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    def test_workers_set_negative_setting(self, mock_server):
        self.config(api_workers=-2, group='api')
        mock_server.return_value = self.server
        self.assertRaises(exception.ConfigInvalid,
                          wsgi_service.WSGIService,
                          'ironic_api')
        self.assertFalse(mock_server.called)

    @mock.patch.object(wsgi_service.wsgi, 'Server', autospec=True)
    @mock.patch('ironic.common.wsgi_service.cheroot_ssl.BuiltinSSLAdapter',
                autospec=True)
    @mock.patch('ironic.common.wsgi_service.validate_cert_paths',
                autospec=True)
    @mock.patch('oslo_service.sslutils.is_enabled', return_value=True,
                autospec=True)
    def test_wsgi_service_with_ssl_enabled(self, mock_is_enabled,
                                           mock_validate_tls,
                                           mock_ssl_adapter,
                                           mock_server):
        self.config(enable_ssl_api=True, group='api')
        self.config(cert_file='/path/to/cert', group='ssl')
        self.config(key_file='/path/to/key', group='ssl')

        mock_server.return_value = self.server

        service_name = 'ironic_api'
        srv = wsgi_service.WSGIService('ironic_api', CONF.api.enable_ssl_api)
        mock_server.assert_called_once_with(server_name=service_name,
                                            wsgi_app=srv.app,
                                            bind_addr=('0.0.0.0', 6385))

        mock_ssl_adapter.assert_called_once_with(
            certificate='/path/to/cert',
            private_key='/path/to/key'
        )
        self.assertIsNotNone(self.server.ssl_adapter)
